# -*- coding: utf8 -*-
from __future__ import unicode_literals

import time

from driverng import Worklist
from driverng.astm import AstmMessageV2, HeaderSegment
from driverng.astm import HprimDate
from driverng.astm import HprimForeign
from driverng.astm import HprimString
from driverng.astm import Segment


class RapidcomSegmentMSH(HeaderSegment):
    sending_application = HprimString(3, default="HIS")
    sending_facility = HprimString(4, default="Hospital")
    receiving_application = HprimString(5, default="Rapidcomm")
    receiving_facility = HprimString(6, default="Hospital")
    message_date = HprimDate(7)
    message_type = HprimString((9, 1), default="ORM")
    event_type = HprimString((9, 2), default="O01")
    control_id = HprimString(10)
    processing_id = HprimString(11, default="P")
    version = HprimString(12, default="2.3")
    accept_ack_type = HprimString(15, default="AL")
    application_ack_type = HprimString(16, default="NE")

    patients = HprimForeign("PID", "header")
    common_order = HprimForeign("ORC", "header")
    queries = HprimForeign("QRD", "header")
    error = HprimForeign("ERR", "header")
    msa = HprimForeign("MSA", "header")


class RapidcomSegmentMSA(Segment):
    ack_code = HprimString(2)
    control_id = HprimString(3)
    flag = HprimString(4)


class RapidcomSegmentPID(Segment):
    medical_record_number = HprimString(4)
    lastname = HprimString((6, 1))
    firstname = HprimString((6, 2))
    middlename = HprimString((6, 3))
    date_of_birth = HprimDate(8, fmt_date="%Y%m%d")
    sex = HprimString(9)
    adress = HprimString(12)
    phone = HprimString(14)
    account_number = HprimString(19)
    social_security_number = HprimString(20)

    patient_visit = HprimForeign("PV1", "patient")
    common_order = HprimForeign("ORC", "patient")


class RapidcomSegmentPV1(Segment):
    patient_class = HprimString(3)
    point_of_care = HprimString((4, 1))
    room = HprimString((4, 2))
    bed = HprimString((4, 3))
    admission_type = HprimString(5)
    attending_physician_id = HprimString((8, 1))
    attending_physician_name = HprimString((8, 2))
    referring_physician_id = HprimString((9, 1))
    referring_physician_name = HprimString((9, 2))
    patient_type = HprimString(19, default="I")
    visit_number = HprimString(20)
    admit_date = HprimString(45)
    discharge_date = HprimString(46)

    common_order = HprimForeign("ORC", "patient")


class RapidcomSegmentORC(Segment):
    order_control = HprimString(2)
    placer_order_number = HprimString((3, 1))
    status = HprimString(6)
    quantity = HprimString((8, 1))
    start_date = HprimString((8, 4))
    end_date = HprimString((8, 5))
    priority = HprimString((8, 6))
    transaction_date = HprimDate(10)
    ordering_physician_id = HprimString((13, 1))
    ordering_physician_name = HprimString((13, 3))

    orders = HprimForeign("OBR", "patient")


class RapidcomSegmentOBR(Segment):
    placer_order_number = HprimString(3)
    filler_order_number = HprimString(4)
    universal_service_identifier = HprimString((5, 1))
    service_identifier_text = HprimString((5, 2))
    collection_identifier = HprimString(11)
    collection_action_code = HprimString(12)
    ordering_physician_id = HprimString((17, 1))
    ordering_physician_name = HprimString((17, 2))
    quantity = HprimString((28, 1))
    start_date = HprimDate((28, 4))
    end_date = HprimDate((28, 5))
    priority = HprimString((28, 6))
    scheduled_date = HprimString(37)

    comments = HprimForeign("NTE", "order")
    results = HprimForeign("OBX", "order")


class RapidcomSegmentOBX(Segment):
    value_type = HprimString(3)
    observation_identifier = HprimString(4)
    value = HprimString(6)
    units = HprimString(7)
    reference_range = HprimString(8)
    abnormal_flags = HprimString(9)
    result_status = HprimString(12)
    observation_date = HprimString(15)
    observer_id = HprimString((17, 1))
    observer_lastname = HprimString((17, 2))
    observer_firstname = HprimString((17, 3))
    equipement_id = HprimString((18, 3))
    model_id = HprimString((18, 4))
    date_analysis = HprimDate(20)

    infos = HprimForeign("NTE", "result")


class RapidcomSegmentQRD(Segment):
    query_date = HprimString(2)
    format_code = HprimString(3, default="R")
    priority = HprimString(4, default="I")
    quantity = HprimString((8, 1), default="1")
    limited_units = HprimString((8, 2), default="RD")
    patient_identifier = HprimString(9)
    medical_record = HprimString(10)


class RapidcomSegmentERR(Segment):
    segment_id = HprimString((2, 1), default="QRD")
    sequence_number = HprimString((2, 2), default="1")
    field_position = HprimString((2, 3), default="8")
    date_of_birth = HprimString((2, 4), default="Patient Not Found")

    queries = HprimForeign("QRD", "queries")


class RapidcomSegmentNTE(Segment):
    set_id = HprimString(2)
    source_comment = HprimString(3)
    comment = HprimString(4)
    comment2 = HprimString(5)


class Driver(AstmMessageV2):
    SPEC = {"code": "rapidcom", "label": "RAPIDCOM HL7"}
    SEGMENT_CLASS_MAP = {
        "MSH": RapidcomSegmentMSH,
        "PID": RapidcomSegmentPID,
        "PV1": RapidcomSegmentPV1,
        "OBR": RapidcomSegmentOBR,
        "OBX": RapidcomSegmentOBX,
        "ORC": RapidcomSegmentORC,
        "NTE": RapidcomSegmentNTE,
        "QRD": RapidcomSegmentQRD,
        "ERR": RapidcomSegmentERR,
        "MSA": RapidcomSegmentMSA,
    }
    DEFAULT_SEPARATORS = "|^~\\&"
    FIRST_SEP_INDEX = 3
    SEGMENT_ENDING = "\r"

    def create(self, **kwargs):
        return RapidcomSegmentMSH(type_segment="MSH", **kwargs)

    def to_string(self, message, separators=None, **kwargs):
        return super(Driver, self).to_string(message, separators, **kwargs) + "\r"

    def encode_worklist(self, worklist, context):
        message = self.create(
            messsage_date=context["now"],
            control_id=context.get("control_id", int(time.time() * 1000)),
        )
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            priority={"A": "S", "S": "S", worklist.ELSE: "R"},
            action={"del": worklist.IGNORE, worklist.ELSE: "NW"},
        ):
            seg_pid = message.patients.add(
                lastname=patient.last_name,
                firstname=patient.first_name,
                middlename=patient.maiden_name,
                date_of_birth=patient.birthdate,
                sex=patient.sex,
            )
            for sample, action, priority in patient.samples:
                seg_pid.medical_record_number = sample.sample_id
                physician = [
                    p for p in patient.physicians if p.type.lower() == "attending"
                ]
                prescriber = [p for p in patient.physicians if p.type == "prescriber"]
                seg_pv1 = seg_pid.patient_visit.add(
                    attending_physician_id=physician[0].code if physician else "",
                    referring_physician_id=prescriber[0].code if prescriber else "",
                    admit_date=sample.request_date,
                )
                seg_orc = seg_pv1.common_order.add(
                    order_control=action,
                    placer_order_number=sample.sample_id,
                    priority=priority,
                    ordering_physician_id=physician[0].code if physician else "",
                )
                for order in sample["orders"]:
                    seg_orc.orders.add(
                        placer_order_number=sample.sample_id,
                        universal_service_identifier=order.code,
                        service_identifier_text=sample.specimen,
                        ordering_physician_id=physician[0].code if physician else "",
                        priority=priority,
                        start_date=sample.request_date,
                    )
        if len(message.patients):
            return message
        if context["query"]:
            for sample_id in [
                seg_qrd.patient_identifier for seg_qrd in context["query"].queries
            ]:
                seg_error = message.error.add()
                seg_error.queries.add()
            return message

    def decode_message(self, message):
        worklist = Worklist(context="result")
        if message.queries:
            worklist = Worklist(context="query")
            for seg_qrd in message.queries:
                worklist.get_or_create(
                    type="patient", sample_id=seg_qrd.patient_identifier
                )
        else:
            # QC
            for seg_orc in message.common_order:
                for seg_obr in seg_orc.orders:
                    if seg_obr.universal_service_identifier == "QC":
                        sample = worklist.get_or_create(type="control")
                        instrument = ""
                        date = None
                        for seg_obx in seg_obr.results:
                            if seg_obx.observation_identifier == "Lot":
                                sample.sample_id = seg_obx.value
                                sample.control_lot = seg_obx.value
                            else:
                                if seg_obx.equipement_id:
                                    instrument = seg_obx.equipment_id
                                if seg_obx.date_analysis:
                                    date = seg_obx.date_analysis
                                sample.results.add(
                                    code=seg_obx.observation_identifier,
                                    value=seg_obx.value.replace(",", "."),
                                    unit=seg_obx.units,
                                    instrument=instrument,
                                    date=date,
                                )
            # Results Patient
            for seg_pid in message.patients:
                for seg_orc in seg_pid.common_order:
                    for seg_obr in seg_orc.orders:
                        sample = worklist.get_or_create(
                            sample_id=seg_obr.placer_order_number, type="patient"
                        )
                        instrument = ""
                        operator = ""
                        date = None
                        for seg_obx in seg_obr.results:
                            if seg_obx.equipement_id:
                                instrument = seg_obx.equipment_id
                            if seg_obx.date_analysis:
                                date = seg_obx.date_analysis
                            if seg_obx.observer_id:
                                operator = seg_obx.observer_id
                            sample.results.add(
                                code=seg_obx.observation_identifier,
                                value=seg_obx.value.replace(",", "."),
                                unit=seg_obx.units,
                                instrument=instrument,
                                operator=operator,
                                date=date,
                            )
        return worklist
